package esercitazione9.graph;

import java.util.ArrayDeque;
import java.util.HashMap;
import java.util.TreeMap;

import esercitazione8.graph.Graph;
import esercitazione8.graph.Node;

public class GraphVisits<T extends Comparable<T>> {

	// Utili per visita schema-DFS
	TreeMap<Node<T>, Integer> discoveryTime;
	TreeMap<Node<T>, Integer> finishTime;
	Integer time;
	

	public Boolean isBipartite(Graph<T> graph, Node<T> u){
		
		Boolean isBipartite = true;
		
		ArrayDeque<Node<T>> queue = new ArrayDeque<Node<T>>();
		HashMap<Node<T>, Colors> colors = new HashMap<Node<T>, Colors>();
		
		queue.add(u);
		colors.put(u, Colors.RED);
		
		Colors currentColor = Colors.RED;
		
		while (!queue.isEmpty() && isBipartite) {
			
			Node<T> a = queue.poll();
			
			currentColor = this.invert(colors.get(a));			
			
			for (Node<T> v : graph.adj(a)) {
			
				if (!colors.containsKey(v)) {
					colors.put(v, currentColor);
					queue.push(v);				
					}
				else if (!colors.get(v).equals(currentColor))
					isBipartite = false;
				
			}
			
			
		}
		
		return isBipartite;
	}
	
	
	private Colors invert(Colors c){
		if (c.equals(Colors.RED))
			return Colors.WHITE;
		else
			return Colors.RED;
	}
	
	
	
	
	public Boolean hasCycles(Graph<T> graph, Node<T> u){
		
		this.discoveryTime = new TreeMap<Node<T>, Integer>();
		this.finishTime = new TreeMap<Node<T>, Integer>();
		this.time = 0;
		
		return this.checkCycles(graph, u);
		
	}
			
	private boolean checkCycles(Graph<T> graph, Node<T> u){
		
		boolean hasCycles = false;
		
		this.time +=1;
		this.discoveryTime.put(u, time);
		
		for (Node<T> v : graph.adj(u)) {
			
			if (!this.discoveryTime.containsKey(v)){
				if (checkCycles(graph, v)) return true;
			}
			else if ((this.discoveryTime.get(v) < this.discoveryTime.get(u)) && !this.finishTime.containsKey(v))
				return true;
		}
		
		this.time +=1;
		this.finishTime.put(u, time);
		
		return hasCycles;
	}
	
	
	
	
	public void printPath(Graph<T> graph, Node<T> u, Node<T> v){
		
		ArrayDeque<Node<T>> queue = new ArrayDeque<Node<T>>();
		HashMap<Node<T>, Node<T>> parents = new HashMap<Node<T>, Node<T>>();
		
		queue.add(u);
		parents.put(u, null);
		
		while (!queue.isEmpty()) {
			
			Node<T> a = queue.poll();
			
			for (Node<T> b : graph.adj(a)) {
			
				if (!parents.containsKey(b)) {
					parents.put(b, a);
					queue.add(b);				
					}
			}
			
		}
		
		System.out.println(parents);
		
		
		ArrayDeque<Node<T>> stack = new ArrayDeque<Node<T>>();
		
		Node<T> p = v;
		
		while ((p != null) && (!p.equals(u))) {
			
			stack.push(p);
			p = parents.get(p);
			
		}
		
		if (p != null) {
			stack.push(p);
			System.out.print(stack);
			}
		else
			System.out.println("Nessun cammino BFS da " + u + " a " + v);
		
	}
	
	
	
}
